#!/usr/bin/env python
# Copyright (c) 2021-2023 Mario S. Könz; License: MIT
import sys


def main() -> None:
    assert len(sys.argv) == 4
    _, name, value, color = sys.argv  # pylint: disable=unbalanced-tuple-unpacking

    # colors from https://docs.gitlab.com/ee/ci/pipelines/settings.html#test-coverage-report-badge
    if color == "green":
        color = "#4c1"
    elif color == "lightgreen":
        color = "#a3c51c"
    elif color == "yellow":
        color = "#dfb317"
    elif color == "red":
        color = "#e05d44"
    elif color == "gray":
        color = "#e05d44"
    else:
        raise NotImplementedError(color)

    print(
        f"""
<svg xmlns="http://www.w3.org/2000/svg" width="116" height="20">
  <linearGradient id="b" x2="0" y2="100%">
    <stop offset="0" stop-color="#bbb" stop-opacity=".1"/>
    <stop offset="1" stop-opacity=".1"/>
  </linearGradient>

  <mask id="a">
    <rect width="116" height="20" rx="3" fill="#fff"/>
  </mask>

  <g mask="url(#a)">
    <path fill="#555" d="M0 0 h62 v20 H0 z"/>
    <path fill="{color} " d="M62 0 h54 v20 H62 z"/>
    <path fill="url(#b)" d="M0 0 h116 v20 H0 z"/>
  </g>

  <g fill="#fff" text-anchor="middle">
    <g font-family="DejaVu Sans,Verdana,Geneva,sans-serif" font-size="11">
      <text x="31" y="15" fill="#010101" fill-opacity=".3">
        {name}
      </text>
      <text x="31" y="14">
        {name}
      </text>
      <text x="89" y="15" fill="#010101" fill-opacity=".3">
        {value}
      </text>
      <text x="89" y="14">
        {value}
      </text>
    </g>
  </g>
</svg>
"""
    )


if __name__ == "__main__":
    main()



# Copyright (c) 2021-2023 Mario S. Könz; License: MIT
# type: ignore
import ruamel.yaml

__all__ = ["yaml", "CommentedMap", "CommentedSeq"]

CommentedMap = ruamel.yaml.comments.CommentedMap
CommentedSeq = ruamel.yaml.comments.CommentedSeq

yaml = ruamel.yaml.YAML()
yaml.indent = 4
yaml.width = 3000
yaml.preserve_quotes = True



# Copyright (c) 2021-2023 Mario S. Könz; License: MIT
import dataclasses as dc
import typing as tp

from ._docker import DockerPayload
from ._docker_executors import DockerRun

__all__ = ["DockerRunPayload"]


@dc.dataclass
class DockerRunPayload(DockerPayload):
    flavor: tp.ClassVar[str] = "docker_run"

    def create_executor(self, parents: tp.Any) -> tp.Any:
        return DockerRun(
            slug=self.auxh.project.slug, services=self.param.services, parents=parents
        )



# Copyright (c) 2021-2023 Mario S. Könz; License: MIT
import io
import shutil
import subprocess
import sys
import typing as tp
from pathlib import Path

from ._components import AllComponents
from ._components import ExtraLevel
from ._components._02_base import BaseComponent
from ._components._03_meta import MetaMixin
from ._components._98_sentinel import SentinelMixin
from ._gitlab import GitlabSetter
from ._proto_namespace import _ProtoNamespace
from ._tick import TickSetter
from ._todo import TodoSetter
from ._util import LazyVersionStr


class CliMixin(  # pylint: disable=too-many-public-methods
    SentinelMixin, MetaMixin, BaseComponent
):
    def __init__(self, target: Path, silent: bool = False) -> None:
        super().__init__()
        self.verbose = not silent
        self.target = Path(target or "devops").resolve()

        if str(Path.cwd()) in str(self.target):
            self.target = self.target.relative_to(Path.cwd())

        valid_suffix = [".yml", ".yaml", ".cfg"]
        candidates = [
            (self.target.parent / "auxilium").with_suffix(sfx) for sfx in valid_suffix
        ]
        self.auxcon_file = candidates[0]

        # fuzzy finder via git root
        if not any(cand.exists() for cand in candidates) and target is None:
            resp = subprocess.run(
                ["git", "rev-parse", "--show-toplevel"],
                capture_output=True,
                check=False,
            )
            if resp.returncode == 0:
                git_root = Path(resp.stdout.decode().strip())
                self.target = git_root / "devops"
                candidates = [
                    (git_root / "auxilium").with_suffix(sfx) for sfx in valid_suffix
                ]

        for cand in candidates:
            if cand.exists():
                self.auxcon_file = cand
                self.load_auxcon()
                break
        self._print(f"adaux {LazyVersionStr()} at your service", fg="blue")

    def demo(self: "CliMixin") -> None:
        out = io.StringIO()
        self.save_auxcon_to_stream(out, level=ExtraLevel.DEMODATA)
        self._print(out.getvalue())

    def show(self: "CliMixin", full: bool = False) -> None:
        out = io.StringIO()
        self.save_auxcon_to_stream(
            out, level=ExtraLevel.ENRICHED if full else ExtraLevel.DEFAULTED
        )
        self._print(out.getvalue())

    def init(  # pylint: disable=too-many-arguments
        self: "CliMixin",
        project_name: str,
        project_slug: str,
        python_version: str,
        author: str,
        force: bool = False,
    ) -> None:
        self.auxcon.project = _ProtoNamespace(
            name=project_name,
            slug=project_slug,
            minimal_version=python_version,
            author=author,
        )
        self.target.mkdir(parents=True, exist_ok=True)

        self.save_auxcon_and_print(ExtraLevel.TEMPLATED, force=force)

    def save_auxcon_and_print(self, level: ExtraLevel, force: bool = False) -> None:
        dest = self.auxcon_file
        over, fg_col = "", "green"
        if dest.exists():
            if not force:
                raise RuntimeError(f"{dest} exists, use -f or --force to overwrite.")
            over, fg_col = "over", "yellow"

        with self.extra(level=level) as aux:
            self.save_auxcon(aux)

        # pylint: disable=protected-access
        self._print(f"{over}written template to {dest}", fg=fg_col)

    def sync(self: "CliMixin") -> None:
        with self.extra(ExtraLevel.FORMATTED) as aux:
            proj = aux.project
        clean: SentinelMixin = self.type_wo_disabled(discard_before="SentinelMixin")()  # type: ignore
        # full holds all possible defaults (in case something got disabled)
        full: SentinelMixin = AllComponents().type_wo_disabled(  # type: ignore
            discard_before="SentinelMixin"
        )()

        for ns in [clean, full]:
            # pylint: disable=no-member
            ns.auxcon.project = _ProtoNamespace(
                name=proj.name,
                slug=proj.slug,
                minimal_version=proj.minimal_version,
                author=proj.author,
            )

        with self.extra(ExtraLevel.FORMATTED) as aux, clean.extra(
            ExtraLevel.TEMPLATED
        ), full.extra(ExtraLevel.TEMPLATED_WITH_NEGATIVE):
            self.update_to_template(clean.auxf, full.auxf)
            self.save_auxcon(aux)

    def bake(self: "CliMixin") -> None:
        if not self.auxcon_file.exists():
            raise RuntimeError(f"{self.auxcon_file} does not exists! use 'aux init'")
        super().bake()

    def gitlab(self: "CliMixin", token: str) -> None:
        with self.extra():
            gls = GitlabSetter(self, token)  # type: ignore
            gls.bake()

    def pipeline(
        self: "CliMixin", token: str, show_success: bool
    ) -> tp.Tuple[bool, bool]:
        with self.extra():
            gls = GitlabSetter(self, token)  # type: ignore
            return gls.pipeline(show_success)

    def tick(
        self: "CliMixin",
        release_message: str,
        major: bool = False,
        minor: bool = False,
        commit: bool = False,
    ) -> None:
        if major and minor:
            raise RuntimeError("cannot set major and minor tick at the same time!")
        if len(release_message) < 16:
            raise RuntimeError("message cannot be shorter than 16 char")
        ticker = TickSetter(self, release_message, major, minor)  # type: ignore
        with self.extra():
            ticker.bake()
            if commit:
                ticker.commit()

    def release(
        self: "CliMixin", token: str, rerelease: bool = False
    ) -> tp.Tuple[bool, bool]:
        with self.extra():
            gls = GitlabSetter(self, token)  # type: ignore
            return gls.release(rerelease)

    def convert(self: "CliMixin", src: Path, dest: Path, force: bool = False) -> None:
        self.auxcon_file = src
        self.load_auxcon()
        self.auxcon_file = dest
        self.save_auxcon_and_print(ExtraLevel.DEFAULTED, force=force)

    def ci(self: "CliMixin", trigger_str: str, dry: bool = False) -> int:
        self._raise_if_disabled("ci")
        with self.extra(ExtraLevel.HYDRATED), self.cwd_to_root():
            return self.run_ci(trigger_str, dry=dry)  # type: ignore

    def cov(self: "CliMixin", open_html: bool = False) -> None:
        self._raise_if_disabled("coverage")
        cov_cache = self.target / "cache" / "coverage"
        datafile = cov_cache / "data_file"
        cov_file = self.target.parent / ".coverage"
        if datafile.exists():
            if cov_file.exists():
                datafile.unlink()
            else:
                datafile.rename(cov_file)

        with self.extra(ExtraLevel.FORMATTED) as aux:
            # pylint: disable=no-member
            cmd = [
                "pytest",
                "--cov",
                f"--cov={aux.project.name}",
                f"--cov-report=html:{cov_cache}",
            ]
            subprocess.run(cmd, check=True)
            cov_file.rename(datafile)

            if open_html:
                cmd = ["open", str(cov_cache / "index.html")]
                subprocess.run(cmd, check=True)

    def dcp(self: "CliMixin", *args: str) -> None:
        self._raise_if_disabled("docker")
        with self.extra() as aux:
            cmd = [
                "docker",
                "compose",
                "-p",
                aux.project.slug,
                "-f",
                str(self.target / "docker" / "compose.yml"),
                *args,
            ]
            subprocess.run(cmd, check=False)

    def run(
        self: "CliMixin", *payload_names: str, force: bool = False, dry: bool = False
    ) -> None:
        self._raise_if_disabled("payload")
        with self.extra(ExtraLevel.HYDRATED), self.cwd_to_root():
            data = self.auxh.payload
            if payload_names == ("$ls",):
                for title, flavor in [
                    ("payloads", "docker_build"),
                    ("with deps", "with_dependency"),
                ]:
                    list_ = list(data.lookup.level(flavor).values())
                    if not list_:
                        continue
                    self._print("=" * len(title))
                    self._print(title)
                    self._print("=" * len(title))
                    for payload in list_:
                        print("-", payload.name)

                return
            try:
                payloads = [data.lookup[x] for x in payload_names]
            except KeyError as err:
                raise RuntimeError(
                    f"payload '{payload_names}' not found! Use aux run ls, or check the {self.auxcon_file.name} file."
                ) from err
            success = self.payload_run(*payloads, force=force, dry=dry)  # type: ignore
            if not success:
                sys.exit(1)

    def graph(
        self: "CliMixin", with_detail: bool = True, with_dependency: bool = True
    ) -> None:
        self._raise_if_disabled("payload")
        with self.extra(ExtraLevel.HYDRATED), self.cwd_to_root():
            self.plot_dependency_graph(with_detail=with_detail, with_dependency=with_dependency)  # type: ignore

    def docs(self: "CliMixin", open_html: bool = False) -> None:
        self._raise_if_disabled("docs")

        docs_cache = self.target / "cache" / "docs"

        with self.extra() as aux:
            # pylint: disable=no-member
            if aux.docs.framework == "sphinx":
                cmd = ["sphinx-apidoc", "-o", aux.docs.root, aux.project.module_dir]
                subprocess.run(cmd, check=True)
                cmd = ["sphinx-build", "-M", "html", aux.docs.root, docs_cache]
                if aux.docs.get("strict", True):
                    cmd.append("-W")
                subprocess.run(cmd, check=True)
                cmd = [
                    "python",
                    self.target / "docs" / "postprocess_html.py",
                    (docs_cache / "html" / aux.project.second_name).with_suffix(
                        ".html"
                    ),
                ]
                subprocess.run(cmd, check=True)
            elif aux.docs.framework == "mkdocs":
                cmd = ["mkdocs", "build", "-f", f"{aux.docs.root}/mkdocs.yml"]
                subprocess.run(cmd, check=True)
            else:
                raise NotImplementedError(aux.docs.framework)

        if open_html:
            cmd = ["open", docs_cache / "html" / "index.html"]
            subprocess.run(cmd, check=True)

    def mp(self: "CliMixin", *args: str) -> None:
        self._raise_if_disabled("mypy")
        cmd = ["git", "add", "-u"]
        subprocess.run(cmd, check=True)
        cmd = [
            "pre-commit",
            "run",
            f"--config={self.target/'pre-commit'/'config.yaml'}",
            "mypy",
            *args,
        ]
        self._run_with_line_limit(cmd)

    def pipi(self: "CliMixin", *args: str) -> None:
        self._raise_if_disabled("pip")
        optional = ""
        if args:
            optional = f'[{",".join(args)}]'
        cmd = [
            "pip",
            "install",
            "-e",
            f"{self.target.parent}{optional}",
            "--config-settings",
            "editable_mode=strict",
        ]
        subprocess.run(cmd, check=False)

    def pl(self: "CliMixin", *args: str) -> None:
        self._raise_if_disabled("pylint")
        cmd = ["git", "add", "-u"]
        subprocess.run(cmd, check=True)
        cmd = [
            "pre-commit",
            "run",
            f"--config={self.target/'pre-commit'/'config.yaml'}",
            "pylint",
            *args,
        ]
        self._run_with_line_limit(cmd)

    def pra(self: "CliMixin", *args: str) -> None:
        self._raise_if_disabled("pre-commit")
        cmd = ["git", "add", "-u"]
        subprocess.run(cmd, check=True)
        if not args:
            args = ("--all",)
        cmd = [
            "pre-commit",
            "run",
            f"--config={self.target/'pre-commit'/'config.yaml'}",
            *args,
        ]
        subprocess.run(cmd, check=False)

    def sdist(self: "CliMixin", zipped: bool = False) -> None:
        self._raise_if_disabled("package")
        cmd = ["python", "setup.py", "sdist"]
        subprocess.run(cmd, check=True)

        if not zipped:
            files = list((self.target.parent / "dist").glob("*.tar.gz"))
            for file in files:
                path = file.with_name(file.name.replace(".tar.gz", ""))
                if path.exists():
                    shutil.rmtree(path, ignore_errors=True)
            cmd = ["tar", "-zxf", *map(str, files), "-C", "dist"]
            subprocess.run(cmd, check=True)
            for file in files:
                file.unlink()

    def todo_and_note(
        self: "CliMixin",
        flavor: str,
        new: tp.Optional[str] = None,
        close: tp.Optional[str] = None,
        gitignore: bool = False,
    ) -> None:
        if new and close:
            raise RuntimeError("cannot set new and close an the same time")

        with self.extra(), self.cwd_to_root():
            todo = TodoSetter(self)  # type: ignore
            if gitignore:
                todo.add_gitignore()
            if new:
                todo.new(new, flavor)
            elif close:
                todo.close(close, flavor)
            else:
                todo.show(flavor)

    def _raise_if_disabled(self, component: str) -> None:
        if not self.is_enabled(component):
            raise RuntimeError(
                f"{component} disabled for this project, check your auxilium file."
            )

    @staticmethod
    def _run_with_line_limit(cmd: tp.List[str], limit: int = 50) -> None:
        proc = subprocess.run(cmd, check=False, stdout=subprocess.PIPE)
        i = 0
        for i, line in enumerate(proc.stdout.decode("utf-8").split("\n")):
            if i < limit:
                print(line)
        if i >= limit:
            print(f"{i-limit} more lines not shown, only the first {limit}...")


class CliRenderer(CliMixin, AllComponents):  # pylint: disable=too-many-ancestors
    pass



# Copyright (c) 2021-2023 Mario S. Könz; License: MIT
import dataclasses as dc
import json
import typing as tp

import requests

__all__ = ["LazyVersionStr"]


class LazyVersionStr:
    def __str__(self) -> str:
        # pylint: disable=import-outside-toplevel,cyclic-import
        import adaux

        return adaux.__version__


@dc.dataclass
class ApiRequestCommunicator:
    base_url: str = dc.field(init=False)
    headers: tp.Dict[str, str] = dc.field(init=False)

    def url(self, *args: str) -> str:
        root = (self.base_url,)
        return "/".join(map(str, root + args))

    def api(self, *args: tp.Any) -> str:
        return self.url("api", "v4", *args)

    def api_request(self, *args: tp.Any, mode: str = "get", **kwgs: tp.Any) -> tp.Any:
        url = self.api(*args)
        if mode == "get":
            assert not kwgs
            return self.get_request(url)
        if mode == "put":
            return self.put_request(url, json=kwgs)
        if mode == "post":
            return self.post_request(url, json=kwgs)
        if mode == "delete":
            return self.delete_request(url)

        raise NotImplementedError(mode)

    def graphql(self, *args: tp.Any) -> str:
        # https://gitlab.com/-/graphql-explorer
        return self.url("api", "graphql", *args)

    def graphql_request(
        self, mode: str = "get", **kwgs: tp.Any
    ) -> tp.Dict[str, tp.Any]:
        url = self.graphql()
        if mode == "get":
            return self.get_request(url, json=kwgs)
        raise NotImplementedError(mode)

    def get_request(self, url: str, **kwgs: tp.Any) -> tp.Dict[str, tp.Any]:
        req = requests.get(url, headers=self.headers, timeout=10, **kwgs)
        if req.status_code != 200:
            raise RuntimeError(f"{req} {req.text}")
        return json.loads(req.text)  # type: ignore

    def put_request(self, url: str, **kwgs: tp.Any) -> tp.Dict[str, tp.Any]:
        req = requests.put(url, headers=self.headers, timeout=10, **kwgs)
        self.assert_with_debug(req, 200)
        return json.loads(req.text)  # type: ignore

    def post_request(self, url: str, **kwgs: tp.Any) -> tp.Dict[str, tp.Any]:
        req = requests.post(url, headers=self.headers, timeout=10, **kwgs)
        self.assert_with_debug(req, 201)
        return json.loads(req.text)  # type: ignore

    def delete_request(self, url: str, **kwgs: tp.Any) -> tp.Dict[str, tp.Any]:
        req = requests.delete(url, headers=self.headers, timeout=10, **kwgs)
        self.assert_with_debug(req, 204)
        return {}

    @classmethod
    def assert_with_debug(cls, req: tp.Any, code: int) -> None:
        if req.status_code != code:
            print(req, req.text)
            assert req.status_code == code



# Copyright (c) 2021-2023 Mario S. Könz; License: MIT
import dataclasses as dc
import functools
import hashlib
import os
import re
import subprocess
import sys
import typing as tp

import requests

from ..._logging import logger


@dc.dataclass(frozen=True)
class DockerShared:
    slug: str
    parents: tp.Tuple["DockerShared", ...] = tuple()

    # Run local registry:
    # docker run -d -p 5000:5000 --name registry registry:2
    # docker container stop registry && docker container rm -v registry
    # export CI_REGISTRY=localhost:5000
    @property
    def registry_host(self) -> str:
        if not self.remote_exists():
            raise RuntimeError(
                "Your run needs a remote registry.\n"
                "In case you have not already, you can launch a local registry with\n"
                "  docker run -d -p 5000:5000 --name registry registry:2\n"
                "and run\n"
                "  export CI_REGISTRY=localhost:5000"
            )
        return os.environ["CI_REGISTRY"]

    def remote_exists(self) -> bool:
        return "CI_REGISTRY" in os.environ

    @property
    def build_deps(self) -> tp.Iterator["DockerBuildMixin"]:
        for parent in self.parents:
            if isinstance(parent, DockerBuildMixin):
                yield parent

    def script(self) -> None:
        # get dependencies
        for job in self.build_deps:
            local_tag = job.local_tag
            if not exists_locally(local_tag):
                if self.remote_exists():
                    remote_tag = job.remote_tag
                    if not exists_locally(remote_tag):
                        cmd = ["docker", "pull", remote_tag]
                        subprocess_run(cmd, check=True, capture_output=True)
                        logger.info("pulled '%s'", remote_tag)

                    tag_image(remote_tag, job.local_tag)

            # I tag every time as it allows different dependencies to
            # map to the same service name without raising issues
            # (useful for local development)
            tag_image(job.local_tag, job.use_tag)

    def _get_tag_parts(self) -> tp.Iterator[tp.Union[bytes, str]]:
        for job in self.build_deps:
            yield job.tag


@dc.dataclass(frozen=True)
class _DockerBuildMixinState:
    service: str
    image_name: str


@dc.dataclass(frozen=True)
class DockerBuildMixin(DockerShared, _DockerBuildMixinState):
    images: tp.List[str] = dc.field(default_factory=list)
    files: tp.List[str] = dc.field(default_factory=list)
    always_build: bool = False

    def script(self) -> None:
        super().script()
        local_tag = self.local_tag
        if self.remote_exists():
            remote_tag = self.remote_tag
            check_and_pull_if_not_existent(remote_tag, local_tag, self.service)
            # download other images which are not a direct dependency (cross-repo)
            for image_tag in self.images:
                image_remote_tag = (
                    f"{self.registry_host}/{self._remove_namespace_local(image_tag)}"
                )
                check_and_pull_if_not_existent(image_remote_tag, image_tag)

        if not exists_locally(self.local_tag) or self.always_build:
            cmd = [
                "docker",
                "--log-level",
                "ERROR",
                "compose",
                "-p",
                self.slug,
                "-f",
                "devops/docker/compose.yml",
                "build",
                self.service,
            ]
            subprocess_run(cmd, check=True)

        # upload
        tag_image(self.use_tag, local_tag)
        if self.remote_exists():
            upload_to_remote(local_tag, remote_tag)
            logger.info("uploaded %s to %s", local_tag, remote_tag)

    @functools.cached_property
    def tag(self) -> str:
        hasher = hashlib.md5()

        for part in self._get_tag_parts():
            if isinstance(part, str):
                hasher.update(part.encode("utf8"))
            else:
                hasher.update(part)

        tag = hasher.hexdigest()[:16]
        return f"aux-{tag}"

    def _get_tag_parts(self) -> tp.Iterator[tp.Union[bytes, str]]:
        yield from super()._get_tag_parts()
        for file in self.files:
            with open(file, "rb") as f:
                yield f.read()

        # we assume the image tag is not moving (i.e. no latest)
        yield from self.images

    @functools.cached_property
    def remote_tag(self) -> str:
        return f"{self.registry_host}/{self.local_tag}"

    @functools.cached_property
    def local_tag(self) -> str:
        image = self.get_image_name()
        return f"{image}:{self.tag}"

    @functools.cached_property
    def use_tag(self) -> str:
        return self.image_name.lower()

    def get_image_name(self) -> str:
        if "CI_PROJECT_NAMESPACE" in os.environ:
            group = os.environ["CI_PROJECT_NAMESPACE"]
            repo = os.environ["CI_PROJECT_NAME"]
            return f"{group}/{repo}/{self.use_tag}".lower()
        return f"{self.use_tag}"

    def _docker_registry_header_and_url(
        self, image: str
    ) -> tp.Tuple[tp.Dict[str, str], str]:
        protocol = "http"
        headers = {}
        if not self.registry_host.startswith("localhost"):
            protocol = "https"
            headers["Authorization"] = f"Bearer {registry_api_token(image)}"

        # https://gitlab.com/gitlab-org/gitlab/-/issues/19470
        url = f"{protocol}://{self.registry_host}/v2"
        return headers, url

    def is_up_to_date(self) -> bool:
        if self.remote_exists():
            return self.exists_remotely()
        return exists_locally(self.local_tag)

    def _remove_namespace_local(self, image: str) -> str:
        if self.registry_host.startswith("localhost"):
            return image.rsplit("/", 1)[1]
        return image

    def exists_remotely(self) -> bool:
        headers, api_url = self._docker_registry_header_and_url(self.get_image_name())
        try:
            req = requests.get(
                api_url + f"/{self.get_image_name()}/tags/list",
                headers=headers,
                timeout=10,
            )
        except requests.ConnectionError as err:
            raise RuntimeError(
                f"are you sure the registry {self.registry_host} is reachable?\n"
                "use: docker run -d -p 5000:5000 --name registry registry:2"
            ) from err

        if req.status_code != 200:
            return False
        tags = req.json()["tags"] or []
        return self.tag in tags

    def tag_and_upload(self, tag: str) -> tp.Tuple[str, str]:
        remote_tag = self.remote_tag
        local_tag = self.local_tag
        assert isinstance(self, DockerBuild)
        check_and_pull_if_not_existent(remote_tag, local_tag)
        release_tag = remote_tag.replace(self.tag, tag)
        upload_to_remote(local_tag, release_tag)
        return local_tag, release_tag

    def pull_if_not_existent(self) -> str:
        remote_tag = self.remote_tag
        local_tag = self.local_tag
        assert isinstance(self, DockerBuild)
        check_and_pull_if_not_existent(remote_tag, local_tag)
        return local_tag


@dc.dataclass(frozen=True)
class _DockerBuildBranchMatchMixinState(_DockerBuildMixinState):
    branch_match: tp.List[tp.Tuple[str, str, str, tp.List[str], tp.Optional[str]]]


@dc.dataclass(frozen=True)
class MatchCommonImpl:
    @classmethod
    def _get_git_branch(cls) -> str:
        branch = os.environ.get("CI_COMMIT_BRANCH", "") + os.environ.get(
            "CI_MERGE_REQUEST_SOURCE_BRANCH_NAME", ""
        )
        if not branch:
            # we are running it locally, and try to get it from git
            proc = subprocess.run(
                ["git", "branch", "--show-current"], check=False, capture_output=True
            )
            if proc.returncode == 0:
                branch = proc.stdout.decode("utf-8").strip()

        assert branch
        return branch


@dc.dataclass(frozen=True)
class DockerBuildBranchMatchMixin(
    MatchCommonImpl, DockerBuildMixin, _DockerBuildBranchMatchMixinState
):
    def _get_tag_parts(self) -> tp.Iterator[tp.Union[bytes, str]]:
        yield from super()._get_tag_parts()

        branch = self._get_git_branch()

        for pkg, var, url, skip, fallback in self.branch_match:
            if branch in skip:
                logger.info("skipping matching branch '%s' for '%s'", branch, pkg)
                continue
            secret = re.search(r"\$(\w+)@", url).group(1)  # type: ignore
            url = url.replace(f"${secret}", os.environ[secret])
            # figure out if matching branch exists

            try_branches = [(branch, "matching")]
            if fallback:
                try_branches.append((fallback, "fallback"))

            for branch, reason in try_branches:
                exists, sha = self._matching_branch_exist(url, branch)
                if exists:
                    logger.info("found %s branch '%s' for '%s'", reason, branch, pkg)
                    yield sha
                    os.environ[var] = f"{pkg}@git+{url}@{branch}"
                    break

    @classmethod
    def _matching_branch_exist(cls, url: str, branch: str) -> tp.Tuple[bool, str]:
        cmd = f"git ls-remote {url} {branch}".split(" ")
        res = subprocess_run(cmd, check=False, capture_output=True)
        if res.returncode != 0:
            raise RuntimeError(res.stderr.decode("utf8"))
        parts = res.stdout.decode("utf8").split()
        if parts:
            sha, refs = parts
            assert branch in refs
            return True, sha
        return False, ""


@dc.dataclass(frozen=True)
class _DockerBuildBaseMatchMixinState(_DockerBuildMixinState):
    base_match: tp.List[tp.Tuple[str, str, tp.List[str], str, str]]


@dc.dataclass(frozen=True)
class DockerBuildBaseMatchMixin(
    MatchCommonImpl, DockerBuildMixin, _DockerBuildBaseMatchMixinState
):
    def _get_tag_parts(self) -> tp.Iterator[tp.Union[bytes, str]]:
        yield from super()._get_tag_parts()
        branch = self._get_git_branch()

        for image_w_ns, var, skip, fallback, default in self.base_match:
            # locally, we do not use the group and repo prefix, hence
            # it needs to be removed
            image = self._remove_namespace_local(image_w_ns)
            if branch in skip:
                logger.info(
                    "skipping matching base '%s' for '%s', using default '%s'",
                    branch,
                    image,
                    default,
                )
                self._fix_environ_and_image(image, var, default)
                yield default
                return

            try_bases = [(fallback, "fallback")]

            for tag, reason in try_bases:
                exists, sha = self._matching_base_exist(image, tag)
                if exists:
                    logger.info("found %s tag '%s' for '%s'", reason, tag, image)
                    self._fix_environ_and_image(image, var, tag)
                    yield sha
                    break
            else:
                logger.info(
                    "no matching base found for '%s', using default '%s'",
                    image,
                    default,
                )
                self._fix_environ_and_image(image, var, default)
                yield tag

    def _fix_environ_and_image(self, image: str, var: str, tag: str) -> None:
        unevaluated = f"{image}:${var}"
        os.environ[var] = tag

        for i, candidate in enumerate(self.images):
            if candidate.endswith(unevaluated):
                self.images[i] = candidate.replace(f"${var}", tag)

    def _matching_base_exist(self, image: str, tag: str) -> tp.Tuple[bool, str]:
        headers, api_url = self._docker_registry_header_and_url(image)
        try:
            headers["Accept"] = "application/vnd.docker.distribution.manifest.v2+json"
            req = requests.get(
                api_url + f"/{image}/manifests/{tag}", headers=headers, timeout=10
            )
            sha = req.json()["config"]["digest"]
            return True, sha
        except (requests.ConnectionError, KeyError):
            return False, ""


@dc.dataclass(frozen=True)
class _DockerRunMixinState:
    services: tp.List[str]


@dc.dataclass(frozen=True)
class DockerRunMixin(DockerShared, _DockerRunMixinState):
    def script(self) -> None:
        super().script()

        cmd = [
            "docker",
            "--log-level",
            "ERROR",
            "compose",
            "-p",
            self.slug,
            "-f",
            "devops/docker/compose.yml",
            "up",
            "--no-build",
            "--no-log-prefix",
            "--abort-on-container-exit",
            *self.services,
        ]
        subprocess_run(cmd, check=True)


@dc.dataclass(frozen=True)
class DockerBuild(
    DockerBuildBaseMatchMixin,
    DockerBuildBranchMatchMixin,
    _DockerBuildBaseMatchMixinState,
    _DockerBuildBranchMatchMixinState,
):
    pass


@dc.dataclass(frozen=True)
class DockerRun(DockerRunMixin, _DockerRunMixinState):
    pass


def subprocess_run(
    *args: tp.Any, **kwgs: tp.Any
) -> "subprocess.CompletedProcess[bytes]":
    sys.stdout.flush()
    try:
        # pylint: disable=subprocess-run-check
        return subprocess.run(*args, **kwgs)
    except subprocess.CalledProcessError as err:
        if err.stdout:
            print("========>> stdout")
            print(err.stdout.decode("utf-8"))
        if err.stderr:
            print("========>> stderr")
            print(err.stderr.decode("utf-8"))
        raise


# try to download for caching (or reuse if already there)
def check_and_pull_if_not_existent(
    remote_tag: str, local_tag: str, service: tp.Optional[str] = None
) -> None:
    if exists_locally(local_tag):
        tag_image(local_tag, remote_tag)
    elif not exists_locally(remote_tag):
        cmd = ["docker", "pull", remote_tag]
        res = subprocess_run(cmd, check=False, capture_output=True)
        if res.returncode != 0:
            return

    tag_image(remote_tag, local_tag)
    if service:
        tag_image(remote_tag, service)


def upload_to_remote(
    local_tag: str, remote_tag: str, registry: tp.Optional[str] = None
) -> None:
    tag_image(local_tag, remote_tag)
    cmd = ["docker", "push", remote_tag]
    if registry:
        cmd += [registry]
    subprocess_run(cmd, check=True, capture_output=True)
    logger.info("pushed '%s'", remote_tag)


def tag_image(src_tag: str, dest_tag: str) -> None:
    cmd = ["docker", "image", "tag", src_tag, dest_tag]
    subprocess_run(cmd, check=True, capture_output=True)
    logger.info("tagged '%s' -> '%s'", src_tag, dest_tag)


def exists_locally(local_tag: str) -> bool:
    cmd = ["docker", "image", "inspect", local_tag]
    res = subprocess_run(cmd, check=False, capture_output=True)
    return res.returncode == 0


@functools.lru_cache
def registry_api_token(image: str) -> str:
    # https://www.pimwiddershoven.nl/entry/request-an-api-bearer-token-from-gitlab-jwt-authentication-to-control-your-private-docker-registry
    # curl --user 'user:pw' 'https://gitlab.com/jwt/auth?client_id=docker&offline_token=true&service=container_registry&scope=repository:administratum/auxilium/adaux-pytest-39:pull,push'
    # curl -H 'Authorization: Bearer token' https://gitlab.com:5050/v2/administratum/auxilium/adaux-pytest-39/tags/list
    server = os.environ["CI_SERVER_URL"]
    url = f"{server}/jwt/auth?client_id=docker&offline_token=true&service=container_registry&scope=repository:{image}:pull,push"
    req = requests.get(
        url,
        auth=(os.environ["CI_REGISTRY_USER"], os.environ["CI_JOB_TOKEN"]),
        timeout=10,
    )
    return req.json()["token"]  # type: ignore



# Copyright (c) 2021-2023 Mario S. Könz; License: MIT
from . import _cli
from . import _cli_mixin
from . import _components
from ._cli import *
from ._components import _00_extra_level
from ._components._00_extra_level import *

__all__ = _cli.__all__ + _00_extra_level.__all__

__version__ = "2.4.1"
